/* 
 * File:   main.c
 * Author: William
 *
 * Created on December 4, 2023, 8:06 PM
 */

// CONFIG1L
#pragma config PLLDIV = 1       // PLL Prescaler Selection bits (No prescale (4 MHz oscillator input drives PLL directly))
#pragma config CPUDIV = OSC1_PLL2// System Clock Postscaler Selection bits ([Primary Oscillator Src: /1][96 MHz PLL Src: /2])
#pragma config USBDIV = 1       // USB Clock Selection bit (used in Full-Speed USB mode only; UCFG:FSEN = 1) (USB clock source comes directly from the primary oscillator block with no postscale)

// CONFIG1H
#pragma config FOSC = XT_XT     // Oscillator Selection bits (XT oscillator (XT))
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable bit (Fail-Safe Clock Monitor disabled)
#pragma config IESO = OFF       // Internal/External Oscillator Switchover bit (Oscillator Switchover mode disabled)

// CONFIG2L
#pragma config PWRT = OFF       // Power-up Timer Enable bit (PWRT disabled)
#pragma config BOR = ON         // Brown-out Reset Enable bits (Brown-out Reset enabled in hardware only (SBOREN is disabled))
#pragma config BORV = 21        // Brown-out Reset Voltage bits (2.1V)
#pragma config VREGEN = OFF     // USB Voltage Regulator Enable bit (USB voltage regulator disabled)

// CONFIG2H
#pragma config WDT = OFF        // Watchdog Timer Enable bit (WDT disabled)
#pragma config WDTPS = 32768    // Watchdog Timer Postscale Select bits (1:32768)

// CONFIG3H
#pragma config PBADEN = ON      // PORTB A/D Enable bit (PORTB<4:0> pins are configured as analog input channels on Reset)
#pragma config LPT1OSC = OFF    // Low-Power Timer 1 Oscillator Enable bit (Timer1 configured for higher power operation)
#pragma config MCLRE = ON       // MCLR Pin Enable bit (MCLR pin enabled; RE3 input pin disabled)

// CONFIG4L
#pragma config STVREN = ON      // Stack Full/Underflow Reset Enable bit (Stack full/underflow will cause Reset)
#pragma config LVP = ON         // Single-Supply ICSP Enable bit (Single-Supply ICSP enabled)
#pragma config BBSIZ = BB1K     // Boot Block Size Select bit (1KW Boot block size)
#pragma config XINST = OFF      // Extended Instruction Set Enable bit (Instruction set extension and Indexed Addressing mode disabled (Legacy mode))

// CONFIG5L
#pragma config CP0 = OFF        // Code Protection bit (Block 0 (000800-001FFFh) or (001000-001FFFh) is not code-protected)
#pragma config CP1 = OFF        // Code Protection bit (Block 1 (002000-003FFFh) is not code-protected)

// CONFIG5H
#pragma config CPB = OFF        // Boot Block Code Protection bit (Boot block (000000-0007FFh) or (000000-000FFFh) is not code-protected)

// CONFIG6L
#pragma config WRT0 = OFF       // Write Protection bit (Block 0 (000800-001FFFh) or (001000-001FFFh) is not write-protected)
#pragma config WRT1 = OFF       // Write Protection bit (Block 1 (002000-003FFFh) is not write-protected)

// CONFIG6H
#pragma config WRTC = OFF       // Configuration Register Write Protection bit (Configuration registers (300000-3000FFh) are not write-protected)
#pragma config WRTB = OFF       // Boot Block Write Protection bit (Boot block (000000-0007FFh) or (000000-000FFFh) is not write-protected)

// CONFIG7L
#pragma config EBTR0 = OFF      // Table Read Protection bit (Block 0 (000800-001FFFh) or (001000-001FFFh) is not protected from table reads executed in other blocks)
#pragma config EBTR1 = OFF      // Table Read Protection bit (Block 1 (002000-003FFFh) is not protected from table reads executed in other blocks)

// CONFIG7H
#pragma config EBTRB = OFF      // Boot Block Table Read Protection bit (Boot block (000000-0007FFh) or (000000-000FFFh) is not protected from table reads executed in other blocks)

#include <xc.h>

#include <stdio.h>
#include <stdlib.h>

// Custom struct to globally save data about the PWM LED settings for use in interrupts
typedef struct {
    uint16_t maxDuty;
    uint16_t brightness;
    int8_t direction;
} PWM_Data;

// Use volatile to prevent compilier optimization
volatile PWM_Data g_pwm;

// 10 bits of resolution
void set_duty(uint16_t duty){
    CCPR1L = (duty >> 2) & 0xFF; // Chop off the first two bits
    DC1B1 = (duty >> 1) & 1; // 2nd bit
    DC1B0 = duty & 1; // 1st bit
}

void __interrupt(high_priority) HighISR(){
    // Add high priority if needed
}

void __interrupt(low_priority) LowISR() {
    // Check for timer0 overflow
    if(TMR0IF){
        LATA1 ^= 1; // Toggle RA1
        
        TMR0IF = 0; // Clear Timer0 interrupt flag
    }
    
    // Check for timer2 overflow
    if(TMR2IF){
        set_duty(g_pwm.brightness);
        
        // Close to top, start counting down, vice versa in else
        if(g_pwm.brightness >= g_pwm.maxDuty - 10){
            g_pwm.direction = -1;
        } else if (g_pwm.brightness <= 10){
            g_pwm.direction = 1;
        }
        
        // Increment or decrement depending on direction
        g_pwm.brightness += 1 * g_pwm.direction;
        
        TMR2IF = 0; // Clear interrupt flag
    }
}

void init_timer0(){
    TMR0L = 0; // Clear out current timer contents
    TMR0H = 0;
    
    // Timer0 settings
    T08BIT = 0; // 16-bit timer
    T0CS = 0; // Use internal clock
    PSA = 0; // Use prescaler
    T0PS2 = 0; // Divide by 8 on the prescaler
    T0PS1 = 1;
    T0PS0 = 0;
    
    // Timer0 interrupt setup
    TMR0IE = 1; // Enable timer0 interrupt
    TMR0IP = 0; // Overflow interrupt is low priority
    
    TMR0ON = 1; // Turn timer ON
}

void init_timer2(){
    // /16 postscale for smoothness (it slows down overflow output)
    T2OUTPS3 = 1;
    T2OUTPS2 = 1;
    T2OUTPS1 = 1;
    T2OUTPS0 = 1;
    
    TMR2IE = 1; // Enable timer2 interrupt
    TMR2IP = 0; // Low priority
    
    TMR2ON = 1; // Start the timer
}

void init_PWM(){
    // Set this value to 0xFF for max resolution, but slowest frequency
    PR2 = 0xFF;
    
    // Set the default PWM settings
    g_pwm.maxDuty = 0x3FF;
    g_pwm.brightness = g_pwm.maxDuty / 4;
    g_pwm.direction = 1;
    
    set_duty(g_pwm.brightness);
    
    // RC2 is the output for PWM
    TRISC2 = 0;
    
    init_timer2();
    
    // Set the CCP module to PWM mode
    CCP1M3 = 1;
    CCP1M2 = 1;
}

int main(int argc, char** argv) {
    // Enable interrupts in high/low priority mode
    IPEN = 1; // Enable priority levels
    GIEH = 1; // Enable high priority interrupts
    GIEL = 1; // Enable low priority interrupts
    
    // RA0 and RA1 are outputs
    TRISA0 = 0;
    TRISA1 = 0;
    
    init_timer0();
    
    init_PWM();

    // Delay using software, while simultaneously blinking other LED using timer
    while(1){
        LATA0 ^= 1;

        // Delay for a while
        for(int i = 0; i < 0xFFFF/2; i++){}
    }

    return (EXIT_SUCCESS);
}


